import os
import re
import subprocess
from tqdm import tqdm


def parse_cli_args(**kwargs):
    """Prepare arguments for command line.

    Args:
        kwargs: dictionary containing option names and values

    Returns:
        List of arguments to pass to subprocess

    Example usage:
      > parse_cli_args(o='output.txt')
      ['-o', 'output.txt']

      > parse_cli_args(f=1, l=2)
      ['-f', '1', '-l', '2']

      > parse_cli_args(h=True)
      ['-h']
    """

    args = []
    for opt, val in kwargs.items():
        # Get option name
        if val:
            args.append('-' + opt)
            if val is not True:
                # Include value if option is not a boolean switch
                args.append(str(val))
    return args


def pdfinfo(pdf_file, **kwargs):
    """Call pdfinfo (Poppler) with optional keyword arguments.

    Args:
        pdf_file: path to pdf file

    Returns:
        Dictionary containing pdf info
    """

    # Parse command line arguments
    args = parse_cli_args(**kwargs)

    # Collect command line arguments
    command_str = ['pdfinfo', '-isodates', *args, pdf_file]

    result = subprocess.Popen(
        command_str, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=1)
    stdout = result.stdout.read().decode()
    stderr = result.stderr.read().decode()

    # Parse output
    result = {}
    for p in stdout.split('\n'):
        if p:
            key, val = p.split(':', 1)
            result[key] = val.strip()

    return result


def pdftotext(pdf_file, **kwargs):
    """Call pdftotext (Poppler) with optional keyword arguments.

    Args:
        pdf_file: path to pdf file

    Returns:
        Extracted text

    Example usage:
        Extract text from 'docuent.pdf' from page 1 to page 10.

        pdftotext('document.pdf', f=1, l=10)
    """

    # Parse command line arguments
    args = parse_cli_args(**kwargs)

    command_str = ['pdftotext', *args, pdf_file, '-']

    result = subprocess.Popen(
        command_str,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=1)
    stdout = result.stdout.read().decode()
    stderr = result.stderr.read().decode()

    if stderr:
        raise ValueError(stderr)
    elif stdout:
        return stdout


def search_string(pattern, string, match_object, context_length):
    """Get details of SRE_Match object.

    Args:
        pattern:        re search pattern
        string:         re search string
        match_object:   SRE_Match object
        context_length: number of words to include before and after

    Returns:
        whole_word: word around pattern match, including prefix and suffix
        context:    words adjacent to pattern match
    """

    idx_1 = match_object.start()
    idx_2 = match_object.end()

    # Get substrings before and after search result
    str_before = string[:idx_1]
    str_after = string[idx_2:]

    # Get actual word match
    str_match_prefix = string[:idx_1].split(' ')[-1:][0]
    str_match_suffix = string[idx_2:].split(' ')[:1][0]
    whole_word = str_match_prefix + pattern + str_match_suffix

    # Get word in context
    words_before = str_before.split(' ')[-1 - context_length:-1]
    words_after = str_after.split(' ')[1:context_length + 1]
    context = ' '.join(words_before + [whole_word] + words_after)

    return whole_word, context


def search_pdf(pdf_name, search_patterns, context_length):
    """Search for text strings inside a pdf.

    Search patterns should be lowercase, unless case-sensitivity is important.

    'ph' will match:
       - photograph
       - PHANTOM
       - pH

    'pH' will match:
       - pH

    Args:
        pdf_name:        path to pdf file
        search_patterns: list of re search patterns
        context_length:  number of words to include before and after

    Returns:
        A list of dictionaries containing search results
    """

    # Convert search patterns to list, if a single search string is provided
    if isinstance(search_patterns, str):
        search_patterns = [search_patterns]

    search_results = []
    pdf_shortname = os.path.split(pdf_name)[1]
    num_pages = int(pdfinfo(pdf_name)['Pages'])

    # Set up tqdm progress bar
    pbar = tqdm(range(num_pages))
    for i in pbar:
        pbar.set_description('Searching {}'.format(pdf_shortname))
        page_num = i + 1

        # Extract raw text on current page with pdftotext
        page_text = pdftotext(pdf_name, f=page_num, l=page_num)

        # Remove unicode control characters
        page_text = re.sub('\r\n', '\n', page_text)
        page_text = re.sub('[\x02\x03\n]', ' ', page_text)

        for pattern in search_patterns:
            if pattern == pattern.lower():
                # Ignore case if pattern is all lowercase
                matches = re.finditer(pattern, page_text, re.IGNORECASE)
            else:
                # Respect case if pattern is mixed or uppercase
                matches = re.finditer(pattern, page_text)

            for match in matches:
                whole_word, context = search_string(
                    pattern, page_text, match, context_length)

                # Update search results
                search_results.append({
                    'document': pdf_name,
                    'page': page_num,
                    'pattern': pattern,
                    'word': whole_word,
                    'context': context,
                })

    return search_results
